from collections.abc import Callable
from pathlib import Path

from parsed_ffmpeg.parse_ffprobe import parse_ffprobe_output

try:
    from tqdm import tqdm
except ImportError:
    _has_tqdm = False
else:
    _has_tqdm = True

from parsed_ffmpeg.ffmpeg import Ffmpeg
from parsed_ffmpeg.types import FfmpegError, FfmpegStatus, FfprobeResult


async def run_ffprobe(
    command: list[str | Path] | str,
    on_error: Callable[[list[str]], None] | None = None,
    on_warning: Callable[[str], None] | None = None,
    raise_on_error: bool = True,
) -> FfprobeResult:
    command_list: list[str] = []
    if isinstance(command, list):
        command_list = [str(part) for part in command]
    elif isinstance(command, str):
        command_list = command.split(" ")
    error_lines: list[str] = []
    std_err_lines: list[str] = []

    def on_error_listener(err: str) -> None:
        error_lines.append(err)

    def on_std_err_listener(line: str) -> None:
        std_err_lines.append(line)

    ffmpeg = Ffmpeg(
        command=command_list,
        on_stderr=on_std_err_listener,
        on_error=on_error_listener,
        on_warning=on_warning,
    )
    await ffmpeg.start()

    if on_error is not None and error_lines:
        on_error(error_lines)
    if raise_on_error and error_lines:
        raise FfmpegError(
            err_lines=error_lines, full_command=command_list, user_command=command
        )

    result = parse_ffprobe_output("\n".join(std_err_lines))
    if len(result.streams) == 0 and result.duration_ms == -1:
        if raise_on_error:
            raise FfmpegError(
                err_lines=std_err_lines,
                full_command=command_list,
                user_command=command
            )
    return result


async def run_ffmpeg(
    command: list[str | Path] | str,
    on_status: Callable[[FfmpegStatus], None] | None = None,
    on_stdout: Callable[[str], None] | None = None,
    on_stderr: Callable[[str], None] | None = None,
    on_error: Callable[[list[str]], None] | None = None,
    on_warning: Callable[[str], None] | None = None,
    overwrite_output: bool = False,
    raise_on_error: bool = True,
    print_progress_bar: bool = False,
    progress_bar_description: str | None = None,
) -> str:
    command_list: list[str] = []
    if isinstance(command, list):
        command_list = [str(part) for part in command]
    elif isinstance(command, str):
        command_list = command.split(" ")
    if overwrite_output and "-y" not in command_list:
        command_list.append("-y")
    if "-progress" in command_list:
        raise ValueError("-progress parameter can't be in command.")
    command_list += ["-progress", "pipe:1"]
    error_lines: list[str] = []
    std_err_lines: list[str] = []

    def on_error_listener(err: str) -> None:
        error_lines.append(err)

    def on_std_err_listener(line: str) -> None:
        std_err_lines.append(line)
        if on_stderr is not None:
            on_stderr(line)

    pbar: tqdm | None = None
    if print_progress_bar and not _has_tqdm:
        raise ImportError(
            "tqdm is not included in your installation of parsed-ffmpeg, so the progress bar can't be used.\n"
            "Include it with `pip install parsed-ffmpeg[tqdm]`"
        )
    if print_progress_bar:
        pbar = tqdm(desc=progress_bar_description, unit="ms")
    try:

        def tqdm_update(status: FfmpegStatus) -> None:
            if on_status:
                on_status(status)
            if not status.out_time_ms or not status.duration_ms or pbar is None:
                return
            pbar.total = int(status.duration_ms)
            pbar.update(int(min(status.out_time_ms, status.duration_ms) - pbar.n))

        ffmpeg = Ffmpeg(
            command=command_list,
            on_status=tqdm_update,
            on_stdout=on_stdout,
            on_stderr=on_std_err_listener,
            on_error=on_error_listener,
            on_warning=on_warning,
        )
        await ffmpeg.start()

        if on_error is not None and error_lines:
            on_error(error_lines)
        if raise_on_error and error_lines:
            raise FfmpegError(
                err_lines=error_lines, full_command=command_list, user_command=command
            )

        return "\n".join(std_err_lines)
    finally:
        if pbar is not None:
            pbar.close()



import os
from dataclasses import dataclass
from enum import StrEnum, auto
from pathlib import Path


class StreamType(StrEnum):
    VIDEO = auto()
    AUDIO = auto()


@dataclass
class BaseStream:
    bitrate_kbs: int
    codec: str
    details: str
    type: StreamType


@dataclass
class AudioStream(BaseStream):
    sample_rate: int
    channels: list[str]


@dataclass
class VideoStream(BaseStream):
    resolution_w: int
    resolution_h: int
    fps: int


@dataclass
class FfprobeResult:
    bitrate_kbs: int
    duration_ms: int
    start_time: float
    streams: list[BaseStream]


@dataclass
class FfmpegStatus:
    frame: int | None = None
    fps: float | None = None
    bitrate: str | None = None
    total_size: int | None = None
    out_time_ms: float | None = None
    dup_frames: int | None = None
    drop_frames: int | None = None
    speed: float | None = None
    progress: str | None = None
    duration_ms: int | None = None
    completion: float | None = None


class FfmpegError(Exception):
    def __init__(
        self,
        err_lines: list[str],
        full_command: list[str],
        user_command: str | list[str | Path],
    ):
        super().__init__("\n".join(err_lines))
        self.err_lines = err_lines
        self.full_command = full_command
        self.user_command = user_command

    def format_error(self) -> str:
        if isinstance(self.user_command, list):
            user_command = (
                f"[{", ".join([f'"{str(part)}"' for part in self.user_command])}]"
            )
        else:
            user_command = self.user_command
        return (
            f"\n\n\tUser command:\n\t\t{user_command}\n"
            f"\tExecuted command:\n\t\t{" ".join([str(part) for part in self.full_command])}\n"
            f"\tWorking directory:\n\t\t{os.getcwd()}\n"
            f"\n{"\n".join(self.err_lines)}"
        )

    def __str__(self) -> str:
        return self.format_error()



import re

from parsed_ffmpeg.types import BaseStream, VideoStream, StreamType, AudioStream, \
    FfprobeResult


def parse_ffprobe_output(output) -> FfprobeResult:
    # Dictionary to store extracted information
    duration_ms: int = -1
    start_time: float = -1
    bitrate_kbs: int = -1
    streams: list[BaseStream] = []

    # Extract video duration and bitrate
    duration_match = re.search(r"Duration: (\S+), start: (\S+), bitrate: (\S+) kb/s",
                               output)
    if duration_match:
        duration_str = duration_match.group(1)
        # Parse the duration (HH:MM:SS.xx)
        hours, minutes, seconds = map(
            float,
            re.split(':', duration_str)
        )
        duration_ms = int(
            hours * 3600000 + minutes * 60000 + seconds * 1000
        )

        start_time = float(duration_match.group(2))
        bitrate_kbs = int(duration_match.group(3))

    # Parse streams
    stream_pattern = re.compile(
        r"Stream #\d+:\d+\[.*?\]\(.*?\): (Video|Audio): (.+?)\s+\((.*?)\), (.+)"
    )
    for stream_match in stream_pattern.finditer(output):
        stream_type = stream_match.group(1)
        stream_codec = stream_match.group(2)
        stream_details = stream_match.group(4)
        if stream_type == 'Video':
            # Extract additional video information
            video_details_match = re.search(
                r"(\d+x\d+).*?, (\d+) kb/s, (\d+) fps", stream_details
            )
            if video_details_match:
                stream_resolution = video_details_match.group(1).split("x")
                stream_bitrate = int(video_details_match.group(2))
                stream_fps = int(video_details_match.group(3))
                streams.append(VideoStream(
                    type=StreamType.VIDEO,
                    details=stream_details,
                    bitrate_kbs=stream_bitrate,
                    codec=stream_codec,
                    resolution_w=int(stream_resolution[0]),
                    resolution_h=int(stream_resolution[1]),
                    fps=stream_fps,
                ))
        elif stream_type == 'Audio':
            # Extract additional audio information
            audio_details_match = re.search(
                r"(\d+) Hz, (.+?), (\d+) kb/s", stream_details
            )
            if audio_details_match:
                stream_sample_rate = int(audio_details_match.group(1))
                stream_channels = audio_details_match.group(2).split(", ")
                stream_bitrate = int(audio_details_match.group(3))
                streams.append(AudioStream(
                    type=StreamType.AUDIO,
                    details=stream_details,
                    channels=stream_channels,
                    bitrate_kbs=stream_bitrate,
                    codec=stream_codec,
                    sample_rate=stream_sample_rate,
                ))

    return FfprobeResult(
        bitrate_kbs=bitrate_kbs,
        duration_ms=duration_ms,
        start_time=start_time,
        streams=streams,
    )
